#include "field.h"

// Divide the numerator and denominator by any common factor
void FieldElement::reduceToLowestTerms() {
  cl_I GCD=gcd(v[0],gcd(v[1],gcd(v[2],gcd(v[3],w))));
  if (GCD<0)
    GCD=-GCD;
  if (GCD!=1){
    v[0]=exquo(v[0],GCD);
    v[1]=exquo(v[1],GCD);
    v[2]=exquo(v[2],GCD);
    v[3]=exquo(v[3],GCD);
    w=exquo(w,GCD);
  }    
}

// construct the element zero
FieldElement::FieldElement(){
  for (int i=0; i<4; i++){
    v[i]=0;
  }
  w=1;
}

// construct the element:  a w^n
FieldElement::FieldElement(cl_I a, int n){
  // Modify n so it is in the range from 0 to 11
  n=n%12;
  if (n<0) {
    n=n+12;
  }
  // if n is in {0,1,2,3} then it is easy
  if ((n>=0) && (n<=3)) {
    v[0]=v[1]=v[2]=v[3]=0;
    v[n]=a;
    w=1;
  } else {
    // we have n>4. We define the answer recursively.
    FieldElement answer;
    answer=FieldElement(a,n-4)*FieldElement(-1,0,1,0);
    // copy the answer over
    v[0]=answer.v[0];
    v[1]=answer.v[1];
    v[2]=answer.v[2];
    v[3]=answer.v[3];
    w=answer.w;
  }   
}

// construct the element: a + b w + c w^2 + d w^3
FieldElement::FieldElement(cl_I a, cl_I b, cl_I c, cl_I d){
  v[0]=a;
  v[1]=b;
  v[2]=c;
  v[3]=d;
  w=1;
}

/******************* read data *******************/

// return the coefficient of w^i (i must be in {0,1,2,3})
cl_I FieldElement::operator[](int i) const {
  return v[i];
}

// return the denominator
cl_I FieldElement::denominator() const {
  return w;
}

/********************* ALGEBRA ********************/

// test for equality
bool FieldElement::operator==(const FieldElement &x) const {
  return ((v[0]==x.v[0]) &&
	  (v[1]==x.v[1]) &&
	  (v[2]==x.v[2]) &&
	  (v[3]==x.v[3]) &&
	  (w==x.w));
}

// copy the data from one field element to another
FieldElement &FieldElement::operator=(const FieldElement &x) {
  v[0]=x.v[0];
  v[1]=x.v[1];
  v[2]=x.v[2];
  v[3]=x.v[3];
  w=x.w;
  return *this;
}

// addition of two field elements
FieldElement FieldElement::operator+(const FieldElement &x) const {
  FieldElement answer;
  answer.w=exquo(w*x.w, gcd(w,x.w));
  for (int i=0; i<4; i++){
    answer.v[i]=v[i]*(exquo(answer.w,w))+x.v[i]*(exquo(answer.w,x.w));
  }
  answer.reduceToLowestTerms();
  return answer;
}

// negation
FieldElement FieldElement::operator-() const {
  FieldElement answer;
  answer.w=w;
  answer.v[0]= -v[0];
  answer.v[1]= -v[1];
  answer.v[2]= -v[2];
  answer.v[3]= -v[3];
  return answer;
}
 
// multiplication  
FieldElement FieldElement::operator*(const FieldElement &x) const {
  // first work with the the numerators
  cl_I coef[8];
  int i,j;
  // initialize the array
  for (i=0; i<8; i++) {
    coef[i]=0;
  }
  // we compute the product of the two polynomials
  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
      coef[i+j]=coef[i+j]+v[i]*x.v[j];
    }
  } 
  // Now coef stores the coeficients of the product polynomial
  FieldElement answer(coef[0],coef[1],coef[2],coef[3]);
  // We know that w^4 = -1 + w^2
  answer.v[0] =answer.v[0]-coef[4];
  answer.v[2] =answer.v[2]+coef[4];
  // We know that w^5 = -w + w^3
  answer.v[1] =answer.v[1]- coef[5];
  answer.v[3] =answer.v[3]+ coef[5];
  // We know that w^6 = -1
  answer.v[0] =answer.v[0]- coef[6];
  // We know that w^7 = -w
  answer.v[1] =answer.v[1]- coef[7];
    
  // now multiply the denominators
  answer.w = w*x.w;

  // reduce
  answer.reduceToLowestTerms();
    
  return answer;
}

// multiplicative inverse
FieldElement FieldElement::inverse() const {
  FieldElement answer=(automorphism5())*
    (automorphism7())*(automorphism11());
  FieldElement norm=operator*(answer);
  // we want to return answer/norm
  // norm is a rational, so we take a short cut:
  answer.v[0]=answer.v[0]*norm.w;
  answer.v[1]=answer.v[1]*norm.w;
  answer.v[2]=answer.v[2]*norm.w;
  answer.v[3]=answer.v[3]*norm.w;
  answer.w=answer.w*norm.v[0];
  answer.reduceToLowestTerms();
  return answer;
}

/********************* field theory ********************/  

// the field automorphism sending w to w^5
FieldElement FieldElement::automorphism5() const {
  FieldElement answer;
  for (int i=0; i<4; i++) {
    answer=answer+FieldElement(v[i],5*i);
  }
  answer.w=w;
  answer.reduceToLowestTerms();
  return answer;
}

// the field automorphism sending w to w^7
FieldElement FieldElement::automorphism7() const {
  FieldElement answer;
  for (int i=0; i<4; i++) {
    answer=answer+FieldElement(v[i],7*i);
  }
  answer.w=w;
  answer.reduceToLowestTerms();
  return answer;
}

// the field automorphism sending w to w^11
FieldElement FieldElement::automorphism11() const {
  FieldElement answer;
  for (int i=0; i<4; i++) {
    answer=answer+FieldElement(v[i],11*i);
  }
  answer.w=w;
  answer.reduceToLowestTerms();
  return answer;
}

/********************* complex numbers ********************/

// real part
FieldElement FieldElement::Re() const {
  FieldElement answer=operator+(conj());
  answer.w=2*answer.w;
  answer.reduceToLowestTerms();
  return answer;
}  

// imaginary part
FieldElement FieldElement::Im() const {
  FieldElement answer=operator-(conj());
  answer.w=2*answer.w;
  answer.reduceToLowestTerms();
  answer=answer*FieldElement(-1,3); // multiply by -i
  return answer;
}

FieldElement::operator cl_N() const {
  cl_N W=cis(pi(float_format(100))/cl_float(6,float_format(100)));
  cl_N ret=cl_float(v[0],float_format(100));
  ret=ret+v[1]*W;
  ret=ret+v[2]*W*W;
  ret=ret+v[3]*W*W*W;
  ret=ret/w;
  return ret;
}


/******************** output *********************/

std::ostream &operator<<(std::ostream &out, const FieldElement &x){
  return out << "(" << x.v[0] << " + " << x.v[1] << " w + " 
	     << x.v[2] << " w^2 + " << x.v[3] << " w^3)/"
	     << x.w;
}
