// ring.h
//
// This file holds information about the field class, which works with
// elements in the ring Z[w] where w is a 12th root of unity.

#ifndef FIELD_H
#define FIELD_H

#include <iostream>
#include <complex>
#include <cln/integer.h>
#include <cln/integer_io.h>
#include <cln/float.h>
#include <cln/complex.h>
using namespace cln;

class FieldElement {
 private:
  // This stores the data for a field element. 
  // If v={1, 2, 3, 4} and w=5, then the element stored is 
  // (1 + 2 w + 3 w^2 + 4 w^3)/5
  // we normalize so that the demoninator is always positive
  cl_I v[4];
  cl_I w;

  // Divide the numerator and denominator by any common factor
  void reduceToLowestTerms();

 public:
  // construct the element zero
  FieldElement();

  // construct the element:  a w^n
  FieldElement(cl_I a, int n);

  // construct the element: a + b w + c w^2 + d w^3
  FieldElement(cl_I a, cl_I b, cl_I c, cl_I d);

  /******************* read data *******************/

  // return the coefficient of w^i (i must be in {0,1,2,3})
  cl_I operator[](int i) const;

  // return the denominator
  cl_I denominator() const;  

  /********************* algebraic operations ********************/

  // test for equality
  bool operator==(const FieldElement &x) const;

  // test for inequality
  inline bool operator!=(const FieldElement &x) const 
    { return ! operator==(x); }

  // copy the data from one field element to another
  FieldElement &operator=(const FieldElement &x);

  // addition of two field elements
  FieldElement operator+(const FieldElement &x) const;

  // negation
  FieldElement operator-() const;

  // subtraction
  inline FieldElement operator-(const FieldElement &x) const
    { return operator+(-x); }

  // multiplication  
  FieldElement operator*(const FieldElement &x) const;

  // multiplicative inverse
  FieldElement inverse() const;

  // division
  inline FieldElement operator/(const FieldElement &x) const 
    { return operator*(x.inverse()); }

  /********************* field theory ********************/
  
  // There are 3 non-trivial field automorphisms.
  // Of course our field is cyclotomic, so the field automorphisms
  // are identified with the integers less that 12 that are relatively
  // prime to 12

  // the field automorphism sending w to w^5
  FieldElement automorphism5() const;

  // the field automorphism sending w to w^7
  FieldElement automorphism7() const;

  // the field automorphism sending w to w^11
  FieldElement automorphism11() const;

  // The norm is the product of all its images
  // under the field automorphism group
  // this should always be a rationl number
  inline FieldElement norm() const 
    { return (*this)*(automorphism5())*
	(automorphism7())*(automorphism11()); }

  /********************* complex numbers ********************/
  
  // we can do some of the things we like with complex numbers
  // especially since i=Sqrt[-1] is in our field

  // complex conjugation
  inline FieldElement conj() const
    { return automorphism11(); }

  // real part
  FieldElement Re() const;

  // imaginary part
  FieldElement Im() const;

  // Because we can compute real and imaginary parts 
  // the Euclidean dot product makes sense.
  inline FieldElement dot(const FieldElement &x) const
    { return (Re()*x.Re())+(Im()*x.Im()); }

  // standard linear algebraic vector projection of a onto b
  // If we think of a and b as vectors, this returns a vector c so that
  // 1) b and c are paralell
  // 2) a-c is perpendicular to b
  FieldElement vectorProjection(const FieldElement &b) const {
    return dot(b)*b/(b*b.conj());
  }
 

  operator cl_N() const;


  /******************** output *********************/

  friend std::ostream &operator<<(std::ostream &out, const FieldElement &x);

};

#endif
